from mindustry_campaign_stats.constants import Planet, SectorNames, ItemIds
from datetime import datetime, timezone
from typing import Dict, List, Union
import dataclasses
import re


@dataclasses.dataclass
class StorageStats:
	capacity: int
	items: Dict[str, int]


@dataclasses.dataclass
class StorageAndProductionStatsMixin:
	storage: StorageStats
	rawProduction: Dict[str, float] # Per minute
	netProduction: Dict[str, float] # Per minute


@dataclasses.dataclass
class SectorStats(StorageAndProductionStatsMixin):
	name: str
	availability: List[str]
	imports: Dict[str, float] # Per minute
	exports: Dict[str, float] # Per minute


@dataclasses.dataclass
class TotalsStats(StorageAndProductionStatsMixin):
	pass


@dataclasses.dataclass
class Stats:
	date: datetime
	planet: Planet
	sectors: Dict[int, SectorStats]
	totals: TotalsStats

	def to_dict(self) -> Dict:
		ret = dataclasses.asdict(self)

		ret['date'] = ret['date'].isoformat()
		ret['planet'] = ret['planet'].value

		return ret


class StatsBuilder:
	settings: Dict[str, Union[bool, float, int, bytes, str]]
	planet: Planet

	sectors_info: Dict[int, Dict]

	def __init__(self, settings: Dict[str, Union[bool, float, int, bytes, str]], planet: Planet):
		self.settings = settings
		self.planet = planet

		self.sectors_info = self.get_sectors_info()

	def build_sectors(self) -> Dict:
		return {
			sector_id: SectorStats(
				name=SectorNames.get(self.planet).get(sector_id, str(sector_id)),
				availability=sector_info.get('resources', []),
				storage=StorageStats(
					capacity=sector_info.get('storageCapacity', 0),
					items=sector_info.get('items', {})
				),
				rawProduction={
					item_id: item_info.get('mean', 0) * 60 for item_id, item_info in sector_info.get('rawProduction', {}).items()
				},
				netProduction={
					item_id: item_info.get('mean', 0) * 60 for item_id, item_info in sector_info.get('production', {}).items()
				},
				imports={
					item_id: item_info.get('mean', 0) * 60 for item_id, item_info in sector_info.get('import', {}).items()
				},
				exports={
					item_id: item_info.get('mean', 0) * 60 for item_id, item_info in sector_info.get('export', {}).items()
				}
			) for sector_id, sector_info in self.sectors_info.items()
		}

	def build_totals(self) -> TotalsStats:
		return TotalsStats(
			storage=StorageStats(
				capacity=sum([
					sector_info.get('storageCapacity', 0) for sector_info in self.sectors_info.values()
				]),
				items={
					item_id: sum([
						sector_info.get('items', {}).get(item_id, 0) for sector_info in self.sectors_info.values()
					]) for item_id in ItemIds.get(self.planet)
				}
			),
			rawProduction={
				item_id: sum([
					sector_info.get('rawProduction', {}).get(item_id, {}).get('mean', 0) * 60 for sector_info in self.sectors_info.values() if sector_info.get('rawProduction', {}).get(item_id, {}).get('mean', 0)
				]) for item_id in ItemIds.get(self.planet)
			},
			netProduction={
				item_id: sum([
					sector_info.get('production', {}).get(item_id, {}).get('mean', 0) * 60 for sector_info in self.sectors_info.values() if sector_info.get('production', {}).get(item_id, {}).get('mean', 0)
				]) for item_id in ItemIds.get(self.planet)
			}
		)

	def get_sectors_info(self) -> Dict[int, Dict]:
		sector_name_regex = re.compile(fr'{self.planet.value}-s-(?P<number>\d+)-info')

		sectors_info = {}

		for key, value in self.settings.items():
			sector_name_match = sector_name_regex.match(key)

			if not sector_name_match:
				continue

			sector_number = int(sector_name_match.groupdict()['number'])

			sectors_info[sector_number] = value

		return sectors_info


def compute(settings: Dict[str, Union[bool, float, int, bytes, str]], planet: Planet) -> Stats:
	builder = StatsBuilder(settings, planet)

	return Stats(
		date=datetime.now(timezone.utc),
		planet=planet,
		sectors=builder.build_sectors(),
		totals=builder.build_totals()
	)



from mindustry_campaign_stats.constants import ItemIds, ItemColors
from mindustry_campaign_stats.stats import Stats
from math import log10, floor
from rich.table import Table
import json


def humanize_number(value, significant_digits=3, strip_trailing_zeros=True):
    powers = [10 ** x for x in (12, 9, 6, 3, 0, -3, -6, -9)]
    human_powers = ['T', 'B', 'M', 'k', '', 'm', 'µ', 'n']
    is_negative = False
    suffix = ''

    if not isinstance(value, float):
        value = float(value)
    if value < 0:
        is_negative = True
        value = abs(value)
    if value == 0:
        decimal_places = max(0, significant_digits - 1)
    elif .001 <= value < 1:
        decimal_places = max(0, significant_digits - int(floor(log10(value))) - 1)
    else:
        p = next((x for x in powers if value >= x), 10 ** -9)
        i = powers.index(p)
        value = value / p
        before = int(log10(value)) + 1
        decimal_places = max(0, significant_digits - before)
        suffix = human_powers[i]

    return_value = ("%." + str(decimal_places) + "f") % value

    if is_negative:
        return_value = "-" + return_value
    if strip_trailing_zeros and '.' in return_value:
        return_value = return_value.rstrip('0').rstrip('.')

    return f'{return_value} {suffix}'


def to_table(computed_stats: Stats) -> Table:
    date = computed_stats.date.astimezone().strftime('%c')

    ret = Table(
        title=f'{date} - {computed_stats.planet.name}',
        show_footer=True
    )

    # Header
    ret.add_column(
        'Sector',
        footer='Totals'
    )

    ret.add_column(
        'Stat',
        footer=f'Storage ({humanize_number(computed_stats.totals.storage.capacity)})\nRaw prod. (/m)\nNet prod. (/m)',
        no_wrap=True
    )

    for item_id in ItemIds.get(computed_stats.planet):
         if computed_stats.totals.storage.items.get(item_id, 0) == 0 and computed_stats.totals.rawProduction.get(item_id, 0) == 0:
             continue

         ret.add_column(
             item_id.replace('-', '\n').title(),
             footer='\n'.join([
                humanize_number(computed_stats.totals.storage.items.get(item_id, 0)),
                humanize_number(computed_stats.totals.rawProduction.get(item_id, 0)),
                humanize_number(computed_stats.totals.netProduction.get(item_id, 0)),
            ]),
             style=ItemColors.get(item_id),
             header_style=ItemColors.get(item_id),
             footer_style=ItemColors.get(item_id),
             no_wrap=True
         )

    # Body
    for sector in sorted(computed_stats.sectors.values(), key=lambda sector: sector.name):
        stat_labels_cell = [
            'Available',
            f'Storage ({humanize_number(sector.storage.capacity)})',
            'Raw prod. (/m)',
            'Net prod. (/m)',
        ]

        if sector.imports:
            stat_labels_cell.append('Imports (/m)')

        if sector.exports:
            stat_labels_cell.append('Exports (/m)')

        row = [
            str(sector.name).replace(' ', '\n'),
            '\n'.join(stat_labels_cell)
        ]

        for item_id in ItemIds.get(computed_stats.planet):
            if computed_stats.totals.storage.items.get(item_id, 0) == 0 and computed_stats.totals.rawProduction.get(item_id, 0) == 0:
                continue

            stat_values_cell = [
                '✓' if item_id in sector.availability else '✕',
                humanize_number(sector.storage.items.get(item_id, 0)),
                humanize_number(sector.rawProduction.get(item_id, 0)),
                humanize_number(sector.netProduction.get(item_id, 0)),
            ]

            if item_id in sector.imports:
                stat_values_cell.append(humanize_number(sector.imports.get(item_id, 0)))

            if item_id in sector.exports:
                stat_values_cell.append(humanize_number(sector.exports.get(item_id, 0)))

            row.append(
                '\n'.join(stat_values_cell)
            )

        ret.add_row(
            *row,
            end_section=True
        )

    return ret


def to_json(computed_stats: Stats, pretty: bool) -> str:
    return json.dumps(
        computed_stats.to_dict(),
        indent=2 if pretty else None,
        separators=None if pretty else (',', ':')
    )



# Original setup.py template: https://github.com/kennethreitz/setup.py

from setuptools import find_packages, setup, Command
from importlib import util as importlib_util
from subprocess import call
from shutil import rmtree
from os import path
import sys

NAME = 'mindustry-campaign-stats'
DESCRIPTION = 'Python API and CLI tool to read Mindustry\'s campaign global stats.'
URL = 'https://github.com/EpocDotFr/mindustry-campaign-stats'
EMAIL = 'contact.nospam@epoc.nospam.fr'
AUTHOR = 'Maxime "Epoc" Gross'
REQUIRES_PYTHON = '>=3.10'
VERSION = None  # Pulled from mindustry_campaign_stats/__version__.py

REQUIRED = [
    'rich~=14.0',
    'mutf8~=1.0',
    'py-ubjson~=0.16',
    'watchdog~=6.0',
]

EXTRAS = {
    'dev': {
        'build~=1.2',
        'twine~=6.0',
        'setuptools>=69',
    },
}

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Operating System :: OS Independent',
    'Environment :: Console',
    'Topic :: Games/Entertainment',
    'Topic :: File Formats',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Intended Audience :: Developers',
]

PROJECT_URLS = {
    'Documentation': 'https://github.com/EpocDotFr/mindustry-campaign-stats?tab=readme-ov-file#usage',
    'Source code': 'https://github.com/EpocDotFr/mindustry-campaign-stats',
    'Issue tracker': 'https://github.com/EpocDotFr/mindustry-campaign-stats/issues',
    'Changelog': 'https://github.com/EpocDotFr/mindustry-campaign-stats/releases',
}

KEYWORDS = ['mindustry', 'settings', 'reader', 'stats']

here = path.abspath(path.dirname(__file__))

try:
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

about = {}

if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")

    spec = importlib_util.spec_from_file_location('__version__', path.join(here, project_slug, '__version__.py'))
    module = importlib_util.module_from_spec(spec)
    spec.loader.exec_module(module)

    about['__version__'] = module.__version__
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')

            rmtree(path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel distribution…')

        call('"{0}" -m build --sdist --wheel'.format(sys.executable), shell=True)

        self.status('Uploading the package to PyPI via Twine…')

        call('twine upload dist/*', shell=True)

        self.status('Pushing git tags…')

        call('git tag v{0}'.format(about['__version__']), shell=True)
        call('git push --tags', shell=True)

        exit()


setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='DBAD',
    entry_points={
        'console_scripts': [
            'mindustry-campaign-stats = mindustry_campaign_stats.cli:cli',
        ]
    },
    classifiers=CLASSIFIERS,
    cmdclass={
        'upload': UploadCommand,
    },
    project_urls=PROJECT_URLS,
    keywords=KEYWORDS
)
